import argparse
import logging
import os

from inspector import __version__
from inspector.app import App, PYWEBVIEW_GUI
from inspector.config import Config

DESC = """\
Command-line entrypoint for Robocorp Inspector, which is
used to create and manage UI locators
"""


def run():
    """Main entrypoint for CLI."""
    parser = argparse.ArgumentParser(
        description=DESC, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "-d",
        "--database",
        metavar="PATH",
        help="path to locators database",
    )
    parser.add_argument(
        "-r",
        "--remote",
        help="remote server for browser management",
    )
    parser.add_argument(
        "-g",
        "--gui",
        choices=PYWEBVIEW_GUI,
        help="used GUI framework",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="be more talkative",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="print version and exit",
    )

    subparsers = parser.add_subparsers(
        title="actions",
        dest="action",
        description="run an action or subset of inspector directly",
        help="name of action",
    )

    add = subparsers.add_parser("add")
    add.add_argument(
        "type",
        choices=["browser", "image", "windows"],
        help="create locator of given type",
    )

    edit = subparsers.add_parser("edit")
    edit.add_argument(
        "name",
        metavar="NAME",
        help="name of locator to edit",
    )

    args = parser.parse_args()

    if args.version:
        print(__version__)
        return

    config = Config()
    config.set("database", args.database)
    config.set("remote", args.remote)
    config.set("gui", args.gui)
    config.set("debug", args.verbose > 0)

    home = config.get("home")
    os.makedirs(home, exist_ok=True)

    logger = logging.getLogger("pywebview")
    for handler in logger.handlers:
        logger.removeHandler(handler)

    log_level = logging.DEBUG if config.get("debug") else logging.INFO
    log_datefmt = "%Y/%m/%d %H:%M:%S"
    log_format = "%(asctime)s.%(msecs)03d › %(levelname)s › %(name)s › %(message)s"
    logging.basicConfig(
        level=log_level,
        format=log_format,
        datefmt=log_datefmt,
        handlers=[
            logging.FileHandler(home / "inspector.log", "w"),
            logging.StreamHandler(),
        ],
    )

    if args.verbose < 2:
        logging.getLogger("PIL").setLevel(logging.INFO)
        logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)
        logging.getLogger("selenium.webdriver.remote.remote_connection").setLevel(
            logging.INFO
        )

    try:
        # Prevent incompatible Qt plugins from affecting linux + conda
        os.environ.pop("QT_QPA_PLATFORM_PLUGIN_PATH")
    except KeyError:
        pass

    app = App(config)
    try:
        if args.action == "add":
            app.add(args.type)
        elif args.action == "edit":
            app.edit(args.name)
        else:
            app.start()
    except KeyboardInterrupt:
        print("User interrupt")
    finally:
        app.stop()


if __name__ == "__main__":
    run()



from inspector.windows.base import Bridge, traceback


class DatabaseMixin(Bridge):
    @traceback
    def names(self):
        return self.ctx.database.names

    @traceback
    def load(self):
        try:
            super().load()
        except AttributeError:
            pass

        name = self.ctx.selected
        if name is None:
            return []

        locator = self.ctx.load_locator(name)
        if locator is None:
            self.logger.error("No locator with name: %s", name)
            return []

        return name, locator

    @traceback
    def save(self, name, locator):
        try:
            super().save()
        except AttributeError:
            pass

        self.ctx.database.update(name, locator)
        self.ctx.force_update()



import os
import platform
import threading
from pathlib import Path
from enum import Enum


IS_WINDOWS = os.name == "nt"


class SupportedInstances(Enum):
    MANAGER = "manager"
    BROWSER = "browser"
    IMAGE = "image"
    RECORDER = "recorder"
    WINDOWS = "windows"


def robocorp_home():
    """Get the absolute path to the user's Robocorp home folder.
    Prefers environment variable ROBOCORP_HOME, if defined.
    """
    env = os.getenv("ROBOCORP_HOME", "")

    if env.strip():
        path = Path(env)
    elif platform.system() == "Windows":
        path = Path.home() / "AppData" / "Local" / "robocorp"
    else:
        path = Path.home() / ".robocorp"

    return path.resolve()


def set_interval(interval):
    """Decorator for calling function every `interval` seconds.
    Starts after first function invocation.
    """
    interval = float(interval)

    def decorator(function):
        def wrapper(*args, **kwargs):
            stop = threading.Event()

            def loop():
                while not stop.wait(interval):
                    function(*args, **kwargs)

            thread = threading.Thread(target=loop)
            thread.daemon = True
            thread.start()
            return stop

        return wrapper

    return decorator



import traceback as tb
import requests
from inspector.webdriver import Webdriver, friendly_name
from inspector.windows.base import Bridge, Window, traceback
from inspector.windows.mixin import DatabaseMixin


class BrowserBridge(DatabaseMixin, Bridge):
    """Javascript API bridge for browser locators."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.elements = []

    @property
    def driver(self):
        return self.ctx.webdriver

    @driver.setter
    def driver(self, value):
        self.ctx.webdriver = value

    @property
    def is_running(self):
        return self.ctx.webdriver is not None and self.ctx.webdriver.is_running

    def status(self):
        try:
            return self.is_running
        except Exception as exc:
            self.logger.exception(exc)
            raise

    def list(self):
        url = self.ctx.config.get("remote")
        if url is None:
            return []

        try:
            response = requests.get(url)
            return response.json()
        except Exception:  # pylint: disable=broad-except
            self.logger.exception(tb.format_exc())
            self.ctx.config.set("remote", None)
            return []

    @traceback
    def connect(self, browser):
        if browser["type"] == "selenium":
            self.logger.info("Connecting to remote webdriver: %s", browser)
            self.driver = Webdriver.from_remote(
                browser["executor_url"],
                browser["session_id"],
                browser["handle"],
            )
        else:
            raise ValueError(f"Unsupported browser type: {browser}")

    @traceback
    def start(self, url=None):
        is_new_session = not self.is_running
        if is_new_session:
            self.driver = Webdriver()
            self.driver.start()

        if url is not None and str(url).strip():
            self.driver.navigate(url)
            response = {"url": url}
        elif is_new_session:
            self.show_guide()
            response = {"url": ""}
        else:
            response = {"url": self.driver.url}

        return response

    @traceback
    def show_guide(self):
        self.driver.show_guide("inspect-guide")

    @traceback
    def stop(self):
        self.driver.stop()

    @traceback
    def pick(self):
        if not self.is_running:
            raise RuntimeError("No active browser session")

        self.driver.clear()
        return self.driver.pick()

    @traceback
    def validate(self, strategy, value):
        if not self.is_running:
            raise RuntimeError("No active browser session")

        try:
            self.driver.clear()
            self.elements = self.driver.find(strategy, value)
            if not self.elements:
                raise ValueError("No matches found")

            screenshot = self.elements[0].screenshot_as_base64
            matches = self.driver.highlight(self.elements)
        except Exception as exc:  # pylint: disable=broad-except
            self.logger.info("Failed to validate: %s", exc)
            self.elements = []
            screenshot = ""
            matches = []

        if not self.driver.url.startswith("data:text/html"):
            source = self.driver.url
        else:
            source = ""

        return {
            "source": source,
            "screenshot": screenshot,
            "matches": [{"name": name, "value": value} for name, value in matches],
        }

    @traceback
    def focus(self, match_id):
        if not self.is_running:
            raise RuntimeError("No active browser session")

        try:
            element = self.elements[int(match_id)]
        except (ValueError, IndexError):
            self.logger.warning("Unexpected highlight index: %s", match_id)
            return

        self.logger.debug("Focusing element #%d: %s", match_id, friendly_name(element))
        self.driver.focus(element)


class BrowserWindow(Window):
    BRIDGE = BrowserBridge
    DEFAULTS = {
        "url": "browser.html",
        "width": 480,
        "height": 720,
        "on_top": True,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._force_closing = False

    def on_closing(self):
        should_close = super().on_closing()
        if not should_close:
            return False

        driver = self._context.webdriver
        if driver is not None and driver.is_running:
            try:
                driver.clear()
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.debug("Failed to clear webdriver: %s", exc)

        return True



from pathlib import Path
from inspector.utils import robocorp_home

_UNSET = object()

HOME = robocorp_home() / "inspector"
PATHS = [Path.cwd() / "inspector.yaml", HOME / "inspector.yaml"]


class Config:
    def __init__(self):
        # TODO: Convert to TypedDict
        self._values = {}
        self.set("home", HOME)

    def get(self, key, default=_UNSET):
        if default is not _UNSET:
            return self._values.get(key, default)

        try:
            return self._values[key]
        except KeyError as err:
            raise KeyError(f"Undefined configuration key: {key}") from err

    def set(self, key, value):
        self._values[key] = value

    def load(self):
        # TODO: Load possible configuration from PATHS
        raise NotImplementedError

    def save(self):
        raise NotImplementedError



import atexit
import os
import sys
from pathlib import Path

from inspector.utils import IS_WINDOWS

from inspector.database import Database
from inspector.telemetry import Telemetry
from inspector.windows import WINDOWS
from inspector.utils import SupportedInstances


class Context:
    INSTANCES = {
        SupportedInstances.MANAGER.value: None,
        SupportedInstances.BROWSER.value: None,
        SupportedInstances.IMAGE.value: None,
        SupportedInstances.RECORDER.value: None,
        SupportedInstances.WINDOWS.value: None,
    }

    def __init__(self, logger, config):
        #: Shared console/file logger
        self.logger = logger
        #: Application configuration
        self.config = config
        #: Created pywebview windows
        self.windows = []
        #: Telemetry client
        self.telemetry = Telemetry(debug=self.is_debug)
        #: Locators database
        self.database = Database(self.config.get("database"))
        #: Currently selected locator (for editing)
        self.selected = None
        #: Active webdriver
        self.webdriver = None

        self.telemetry.start_worker()
        atexit.register(self._on_exit)

    def _on_exit(self):
        if self.webdriver is not None:
            self.webdriver.stop()

    @property
    def is_debug(self):
        return self.config.get("debug")

    @property
    def entrypoint(self):
        # note: pywebview uses sys.argv[0] as base
        base = Path(sys.argv[0]).resolve().parent
        static = Path(__file__).resolve().parent / "static"
        return os.path.relpath(str(static), str(base))

    def open_window(self, kind, **kwargs):
        self.logger.debug("Opening window: %s", kind)

        if kind == "windows" and not IS_WINDOWS:
            raise NotImplementedError(
                "Windows locators are not available on non Windows platforms"
            )

        factory = WINDOWS.get(kind)
        if factory is None:
            raise KeyError(f"Unknown window type: {kind}")

        window = self.INSTANCES[kind]
        if window is not None and window.is_valid:
            window.restore()
            return window

        window = factory.create(self, **kwargs)
        self.INSTANCES[kind] = window

        self.windows.append(window)
        return window

    def close_windows(self):
        for window in reversed(self.windows):
            if window.is_valid:
                self.logger.debug("Closing window: %s", window)
                window.destroy()

    def force_update(self):
        manager = self.INSTANCES["manager"]
        if manager is not None:
            manager.evaluate_js("window.pywebview.state.update()")

    def load_locator(self, name):
        try:
            with self.database.lock:
                self.database.load()
                return self.database.get(name)
        except KeyError:
            return None



import logging
import os
import platform
import traceback as tb
from abc import ABC
from enum import Enum, auto
from functools import wraps
from typing import Callable
from threading import Timer

import webview  # type: ignore


def traceback(method: Callable):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        try:
            self.logger.debug(f"{type(self).__name__}.{method.__name__}()")
            return method(self, *args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            self.logger.exception(tb.format_exc())
            raise

    return wrapper


APP_WINDOW_TITLE = "UI Inspector"


class WindowState(Enum):
    CREATED = auto()
    SHOWN = auto()
    LOADED = auto()
    CLOSING = auto()
    CLOSED = auto()


class Bridge(ABC):
    # pylint: disable=too-few-public-methods
    def __init__(self, context):
        self.logger = logging.getLogger(__name__)
        self._window = None  # Injected after window creation
        self.ctx = context

    @property
    def window(self):
        return self._window

    def set_window(self, window):
        self._window = window

    @traceback
    def close(self):
        if self.window:
            self.window.close()


class Window(ABC):
    BRIDGE = Bridge
    DEFAULTS = {
        "title": "Title",
        "url": "index.html",
        "width": 640,
        "height": 480,
    }

    @classmethod
    def create(cls, context, **kwargs):
        options = dict(cls.DEFAULTS)
        options.update(kwargs)
        options.setdefault("title", APP_WINDOW_TITLE)
        options["url"] = os.path.join(context.entrypoint, options["url"])

        bridge = cls.BRIDGE(context)
        window = webview.create_window(js_api=bridge, **options)

        instance = cls(context, window, bridge)
        bridge.set_window(instance)

        return instance

    def __init__(self, context, window, bridge):
        self._context = context
        self._window = window
        self._bridge = bridge
        self._state = WindowState.CREATED

        # Attach event callbacks
        self._window.events.closed += self.on_closed
        self._window.events.closing += self.on_closing
        self._window.events.shown += self.on_shown
        self._window.events.loaded += self.on_loaded

        # MacOS closing flag
        self._force_closing = False

    def __getattr__(self, name):
        return getattr(self._window, name)

    @property
    def logger(self):
        return self._context.logger

    @property
    def state(self):
        return self._state

    @property
    def is_valid(self):
        return self.state not in (WindowState.CLOSING, WindowState.CLOSED)

    def close(self):
        Timer(0.2, self.destroy).start()

    def on_shown(self):
        self._state = WindowState.SHOWN

    def on_loaded(self):
        self._state = WindowState.LOADED

    def on_closing(self):
        self._state = WindowState.CLOSING

        if self._force_closing:
            return True
        elif platform.system() == "Darwin":
            self.logger.debug("Force closing window")
            self._force_closing = True
            self.close()
            return False
        else:
            return True

    def on_closed(self):
        self._state = WindowState.CLOSED



from typing import Dict

import base64
import logging
import os
import platform
import traceback
from pathlib import Path, WindowsPath
from urllib.parse import urlparse

from selenium.webdriver import ChromeOptions, Remote  # type: ignore
from selenium.common.exceptions import (  # type: ignore
    InvalidSessionIdException,
    JavascriptException,
    NoSuchWindowException,
    TimeoutException,
    WebDriverException,
    StaleElementReferenceException,
)

from RPA.core import webdriver  # type: ignore


def load_resource(filename):
    path = Path(__file__).parent / "static" / "resources" / str(filename)
    with open(path, "r", encoding="utf-8") as file:
        return file.read()


def chrome_options():
    opts = ChromeOptions()

    preferences = {
        "safebrowsing.enabled": True,
        "credentials_enable_service": False,
        "profile.password_manager_enabled": False,
    }

    opts.add_argument("--no-sandbox")
    opts.add_argument("--allow-running-insecure-content")
    opts.add_argument("--disable-dev-shm-usage")
    opts.add_argument("--disable-web-security")
    opts.add_experimental_option("prefs", preferences)
    opts.add_experimental_option(
        "excludeSwitches", ["enable-logging", "enable-automation"]
    )

    return opts


def friendly_name(element):
    tag = element.tag_name.lower()

    if tag in ("h1", "h2", "h3", "h4", "h5", "h6", "a"):
        return element.text

    if tag == "img":
        alt = element.get_attribute("alt")
        if alt.strip():
            return alt

        url = element.get_attribute("src")
        name = os.path.basename(urlparse(url).path)
        if not url.startswith("data:image") and name.strip():
            return name

        return "Image"

    # TODO: Add div logic
    # TODO: Add input type and title logic

    # Human-friendly names for non-descriptive tags,
    # expand as necessary
    names = {
        # Navigation
        "nav": "Navigation",
        "input": "Input",
        # Text
        "b": "Bold",
        "i": "Italic",
        "br": "Line Break",
        "p": "Paragraph",
        "pre": "Preformatted Text",
        "samp": "Sample Text",
        # Lists
        "li": "List Item",
        "ol": "Ordered List",
        "ul": "Unordered List",
        # Tables
        "tbody": "Table Rows",
        "th": "Table Header",
        "thead": "Table Header",
        "tfoot": "Table Footer",
        "tr": "Table Row",
        "col": "Table Column",
        "td": "Table Cell",
        "hgroup": "Heading Group",
        "colgroup": "Column Group",
        # Misc
        "hr": "Horizontal Line",
    }

    if tag in names:
        return names[tag]

    return tag


class WebdriverError(Exception):
    """Common exception for all webdriver errors."""


class Webdriver:
    SCRIPT_TIMEOUT = 30.0  # seconds

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.selenium = None
        self.resources = {}
        self.is_remote = False
        self._load_resources()

    @classmethod
    def from_remote(cls, executor_url, session_id, handle):
        # Override new session command to prevent spawning windows
        execute = Remote.execute

        def _execute_patched(self, driver_command, params=None):
            if driver_command == "newSession":
                return {"success": 0, "value": None, "sessionId": session_id}
            else:
                return execute(self, driver_command, params)

        try:
            Remote.execute = _execute_patched
            driver = Remote(
                command_executor=executor_url,
                desired_capabilities={},
            )
            driver.session_id = session_id
            driver.switch_to_window(handle)
        finally:
            Remote.execute = execute

        instance = cls()
        instance.selenium = driver
        instance.is_remote = True

        return instance

    @property
    def is_running(self):
        if not self.selenium:
            return False

        try:
            try:
                # Mock interaction to check if webdriver is still available
                _ = self.selenium.current_window_handle
            except NoSuchWindowException:
                # Handle tabs closed by user
                handles = self.selenium.window_handles
                self.selenium.switch_to_window(handles[-1])
            return True
        except Exception:  # pylint: disable=broad-except
            self.selenium = None
            return False

    @property
    def title(self):
        if not self.selenium:
            return None

        return self.selenium.title

    @property
    def url(self):
        if not self.selenium:
            return None

        return self.selenium.current_url

    def _load_resources(self):
        self.resources = {
            "style": load_resource("inspector.css"),
            "inspector": load_resource("inspector.js"),
            "recorder": load_resource("recorder.js"),
            "inspect-guide": load_resource("guide.html"),
            "recording-guide": load_resource("browser-recording-guide.html"),
        }

    def _exec_resources(self):
        if not self.selenium.find_elements_by_id("inspector-style"):
            self.selenium.execute_script(self.resources["inspector"])
            self.selenium.execute_script(self.resources["recorder"])
            self.selenium.execute_script(
                'var style = document.getElementById("inspector-style");'
                f'var content = document.createTextNode(`{self.resources["style"]}`);'
                "style.appendChild(content);"
            )

    def _exec_driver(self, func, *args, **kwargs):
        def error(msg, *error_args):
            self.logger.warning(msg, error_args)
            raise WebdriverError(msg)

        if not self.selenium:
            error("No available webdriver")

        try:
            self._exec_resources()
            return func(*args, **kwargs)
        except TimeoutException:
            error("Timeout while running script")
        except JavascriptException as exc:
            error("Error while running script: %s", exc)
        except WebdriverError as exc:
            error("Webdriver error: %s", exc)
        except (InvalidSessionIdException, NoSuchWindowException) as exc:
            self.selenium = None
            error(exc)

    def _exec_script(self, script, *args, **kwargs):
        return self._exec_driver(self.selenium.execute_script, script, *args, **kwargs)

    def _exec_async_script(self, script, *args, **kwargs):
        return self._exec_driver(
            self.selenium.execute_async_script, script, *args, **kwargs
        )

    def _init_driver(self):
        browsers = webdriver.DRIVER_PREFERENCE[platform.system()]
        for browser in browsers:
            if browser == "Chrome":
                kwargs = {"chrome_options": chrome_options()}
            else:
                kwargs = {}

            def _create_driver(path=None):
                # pylint: disable=cell-var-from-loop
                if path is not None:
                    if isinstance(path, WindowsPath):
                        path = str(path.resolve())
                    return webdriver.start(browser, executable_path=path, **kwargs)
                else:
                    return webdriver.start(browser, **kwargs)

            try:
                # Try to use webdriver already in cache
                cache = webdriver.cache(browser)
                if cache:
                    try:
                        return _create_driver(cache)
                    except Exception:  # pylint: disable=broad-except
                        pass

                # Try to download webdriver
                download = webdriver.download(browser)
                if download:
                    return _create_driver(download)

                # No webdriver required
                return _create_driver()
            except WebDriverException as exc:
                self.logger.warning("Failed to start '%s': %s", browser, exc)
            except Exception:  # pylint: disable=broad-except
                self.logger.error("Unhandled exception:\n%s", traceback.format_exc())

        raise ValueError("No valid browser found")

    def _to_finder(self, strategy):
        strategy = str(strategy).lower()
        finder = {
            "class": self.selenium.find_elements_by_class_name,
            "css": self.selenium.find_elements_by_css_selector,
            "id": self.selenium.find_elements_by_id,
            "link": self.selenium.find_elements_by_link_text,
            "name": self.selenium.find_elements_by_name,
            "tag": self.selenium.find_elements_by_tag_name,
            "xpath": self.selenium.find_elements_by_xpath,
        }[strategy]

        if not finder:
            raise ValueError(f"Unknown search strategy: {strategy}")

        return finder

    def start(self):
        if self.selenium:
            self.logger.warning("Webdriver already running")

        self.logger.debug("Starting browser")
        self.selenium = self._init_driver()
        self.selenium.set_script_timeout(self.SCRIPT_TIMEOUT)

    def stop(self):
        if not self.selenium:
            return

        if self.is_remote:
            self.logger.debug("Skipping close for remote browser")
            return

        self.logger.debug("Stopping browser")
        self.selenium.quit()
        self.selenium = None

    def show_guide(self, guide_resources: str):
        guide = self.resources[guide_resources].encode("utf-8")
        payload = base64.b64encode(guide).decode("utf-8")
        self.selenium.get(f"data:text/html;base64,{payload}")

    def navigate(self, url):
        self.selenium.get(url)

    def pick(self):
        def picker():
            locators = self.selenium.execute_async_script(
                "var callback = arguments[arguments.length - 1];"
                "Inspector.startPicker(callback);"
            )

            if not locators:
                return {}

            options = {}
            elements = set()
            for name, value in locators:
                strategy = name.split(":", 1)[0]
                finder = self._to_finder(strategy)
                matches = finder(value)

                if len(matches) == 1:
                    options[name] = {"strategy": strategy, "value": value}
                    elements.add(matches[0])

            if len(elements) > 1:
                # TODO: Inform user somehow? How to test?
                self.logger.error("Picker options matching multiple elements")

            return options

        self.logger.debug("Starting interactive picker")
        return self._exec_driver(picker)

    def find(self, strategy, value):
        self.logger.debug("Finding elements: %s:%s", strategy, value)
        finder = self._to_finder(strategy)
        return self._exec_driver(finder, value)

    def highlight(self, elements):
        try:
            self.logger.debug("Highlighting %d element(s)", len(elements))
            names = [friendly_name(element) for element in elements]
            values = self._exec_async_script(
                (
                    "var callback = arguments[arguments.length - 1];"
                    "var elements = arguments[0];"
                    "var tags = Inspector.describeElements(elements);"
                    "Inspector.highlightElements(elements);"
                    "callback(tags);"
                ),
                elements,
            )
            return list(zip(names, values))
        except StaleElementReferenceException as exc:
            self.logger.warning(exc)
            return []

    def focus(self, element):
        try:
            self._exec_script(
                "var element = arguments[0]; Inspector.focusElement(element);",
                element,
            )
        except StaleElementReferenceException as exc:
            self.logger.warning(exc)

    def clear(self):
        self.logger.debug("Clearing highlights")
        return self._exec_script("Inspector.removeHighlights();")

    def record_event(self) -> Dict:
        self._exec_resources()

        result = self.selenium.execute_async_script(
            "var callback = arguments[arguments.length - 1];"
            "Recorder.recordEvent(callback);"
        )
        self.logger.debug("Selenium execute async script finished, result: %s", result)
        return result

    def stop_recording(self) -> None:
        self.selenium.execute_script("Recorder.stop();")



from typing import List, Optional

from typing_extensions import Literal, TypedDict

from selenium.common.exceptions import (  # type: ignore
    JavascriptException,
    TimeoutException,
)
from inspector.windows.base import Window, traceback
from inspector.windows.browser import BrowserBridge


class AppendMessage(TypedDict):
    value: str
    type: str
    trigger: Literal["click", "change", "unknown"]


class RecordEvent(TypedDict):
    list: List[AppendMessage]
    actionType: Literal["exception", "stop", "append"]
    url: Optional[str]


class TabInfo(TypedDict):
    title: str
    url: str


class RecorderBridge(BrowserBridge):
    """Javascript API bridge for recorder functionality."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._current_url: Optional[str] = None
        self._messages: List[AppendMessage] = []

    @traceback
    def start_recording(self) -> List[AppendMessage]:
        # TODO: Handle as async in future instead of blocking
        self._record()
        self.logger.debug("Recorded events: %s", self._messages)
        return self._messages

    def _record(self):
        self._messages = []
        while True:
            # TODO: Could we try catch the page change?
            try:
                event: RecordEvent = self.driver.record_event()
            except JavascriptException as exc:
                self.logger.debug("Ignoring Javascript exception: %s", exc)
                event: RecordEvent = {
                    "actionType": "exception",
                    "list": [],
                    "url": self._current_url,
                }
            except TimeoutException:
                self.logger.debug("Retrying after script timeout")
                event: RecordEvent = {
                    "actionType": "exception",
                    "list": [],
                    "url": self._current_url,
                }

            if not event:
                self.logger.error("Received empty event: %s", event)
                continue

            if not self._handle_event(event):
                return

    def _handle_event(self, event: RecordEvent) -> bool:
        event_type = event["actionType"]
        event_url = event["url"]

        if not self._current_url:
            self._current_url = event_url
        elif event_url != self._current_url:
            message: AppendMessage = {
                "type": "comment",
                "value": f"Recorder detected that URL changed to {event_url}",
                "trigger": "unknown",
            }
            self._current_url = event_url
            self._messages += [message]

        if event_type == "exception":
            return True  # Ignore errors for now
        elif event_type == "event":
            self.logger.debug("Received event(s) from page: %s", event["list"])
            self._messages += event["list"]
            return True
        elif event_type == "stop":
            self.logger.debug("Received stop from page")
            self.driver.stop_recording()
            self._messages += event["list"]
            return False
        else:
            raise ValueError(f"Unknown event type: {event_type}")

    @traceback
    def show_guide(self):
        self.driver.show_guide("recording-guide")


class RecorderWindow(Window):
    BRIDGE = RecorderBridge
    DEFAULTS = {
        "url": "recorder.html",
        "width": 480,
        "height": 720,
        "on_top": True,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._force_closing = False

    def on_closing(self):
        should_close = super().on_closing()
        if not should_close:
            return False

        driver = self._context.webdriver
        if driver is not None and driver.is_running:
            try:
                driver.clear()
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.debug("Failed to clear webdriver: %s", exc)

        return True



from inspector.windows.base import Bridge, Window, traceback
from inspector.windows.mixin import DatabaseMixin
from inspector.windows_driver import (
    WindowsDriver,
    WindowsLocator,
    MatchedWindowsLocators,
    OpenWindows,
)


class WindowsBridge(DatabaseMixin, Bridge):
    """Javascript API bridge for windows locators."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.windows_driver = WindowsDriver()

    @traceback
    def pick(self, active_window: str) -> WindowsLocator:
        self.logger.info("Picking windows locator from=%s", active_window)
        return self.windows_driver.listen(active_window)

    @traceback
    def validate(self, active_window: str, value: str) -> MatchedWindowsLocators:
        self.logger.info("Validate locator=%s - %s", active_window, value)
        return self.windows_driver.validate(active_window, value)

    @traceback
    def list_windows(self) -> OpenWindows:
        self.logger.info("List available windows")
        return self.windows_driver.list_windows()

    @traceback
    def focus(self, active_window: str, value: str) -> None:
        self.logger.info("Focus element")
        self.windows_driver.focus(active_window, value)


class WindowsWindow(Window):
    BRIDGE = WindowsBridge
    DEFAULTS = {
        "url": "windows.html",
        "width": 560,
        "height": 720,
    }



import logging
import sys
import webview  # type: ignore
from inspector.context import Context
from inspector.metric import MetricStart

PYWEBVIEW_GUI = ["qt", "gtk", "cef", "mshtml", "edgechromium", "edgehtml"]


class App:
    def __init__(self, config):
        self.logger = logging.getLogger(__name__)
        self.ctx = Context(self.logger, config)

    def start(self, root="manager"):
        assert not self.ctx.windows, "Application already running"

        self.ctx.database.load()
        if self.ctx.database.error:
            self.logger.warning(*self.ctx.database.error)

        root = self.ctx.open_window(root)
        root.closed += self.stop

        self.ctx.telemetry.send(MetricStart())
        webview.start(self._on_start, debug=self.ctx.is_debug)

    def stop(self):
        self.ctx.close_windows()

    def add(self, kind):
        self.start(kind)

    def edit(self, name):
        locator = self.ctx.load_locator(name)

        if locator is None:
            names = "\n".join(f" - {name}" for name in self.ctx.database.names)
            if names:
                print(f"No locator with name: {name}\n")
                print(f"Possible options:\n{names}")
            else:
                print("No locators in database")
            sys.exit(1)

        self.ctx.selected = name
        self.start(locator["type"])

    def _on_start(self):
        self.logger.info("Starting root window")
        if self.ctx.is_debug:
            # Print URL for automation
            root = self.ctx.windows[0]
            self.logger.info(root.get_current_url())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inspector', 'inspector.windows']

package_data = \
{'': ['*'], 'inspector': ['static/*', 'static/resources/*']}

install_requires = \
['mss>=6.1.0,<7.0.0',
 'pynput-robocorp-fork>=4.0.0,<5.0.0',
 'pywebview==3.6.3',
 'requests>=2.25.1,<3.0.0',
 'rpaframework-core>=7.0.1,<8.0.0',
 'rpaframework-recognition>=2.0.0,<3.0.0',
 'typing-extensions>=3.10.0,<4.0.0',
 'uiautomation>=2.0.16,<3.0.0']

extras_require = \
{':python_full_version < "3.7.6" and sys_platform == "win32" or python_full_version > "3.7.6" and python_full_version < "3.8.1" and sys_platform == "win32" or python_full_version > "3.8.1" and sys_platform == "win32"': ['pywin32>=302,<304']}

entry_points = \
{'console_scripts': ['inspector = inspector.cli:run']}

setup_kwargs = {
    'name': 'robocorp-inspector',
    'version': '0.6.4',
    'description': 'Robocorp Inspector',
    'long_description': '# Robocorp Inspector\n\nRobocorp Inspector is a tool for exploring various user interfaces\nand developing ways to target elements within them. An expression\nthat can target specific UI elemements is called a _locator_, and\nthese locators can be used to automate applications typically\nused by humans.\n\n## Development\n\nThe project uses `invoke` for overall project management, `poetry` for\npython dependencies and environments, and `yarn` for Javascript dependencies\nand building.\n\nBoth `invoke` and `poetry` should be installed via pip: `pip install poetry invoke`\n\n- To see all possible tasks: `invoke --list`\n- To run the project: `invoke run `\n\nAll source code is hosted on [GitHub](https://github.com/robocorp/inspector/).\n\n## Usage\n\nRobocorp Inspector is distributed as a Python package with all front-end\ncomponents compiled and included statically.\n\nIf the package (and all required dependencies) is installed manually,\nit can be run with the command: `inspector`.\n\n---\n\n<p align="center">\n  <img height="100" src="https://cdn.robocorp.com/brand/Logo/Dark%20logo%20transparent%20with%20buffer%20space/Dark%20logo%20transparent%20with%20buffer%20space.svg">\n</p>\n',
    'author': 'Ossi Rajuvaara',
    'author_email': 'ossi@rajuvaara.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/robocorp/inspector',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# pylint: disable=cell-var-from-loop
import base64
import os
import logging
from pathlib import Path
from threading import Lock
from typing import Optional, Dict

from RPA.core.locators import (  # type: ignore
    LocatorsDatabase,
    Locator,
    BrowserDOM,
    ImageTemplate,
    sanitize_name,
)


class DatabaseError(Exception):
    """Common exception for Inspector's database errors."""


class Database:
    """Common interface for locators database."""

    def __init__(self, path: Optional[str] = None):
        self.logger = logging.getLogger(__name__)
        self.cache: Dict[str, Dict] = {}
        self.lock = Lock()

        path = str(Path(path)) if path is not None else path
        self.database = LocatorsDatabase(path)

    @property
    def path(self):
        return self.database.path

    @property
    def error(self):
        return self.database.error

    @property
    def names(self):
        return list(sorted(self.database.locators.keys()))

    def _create_cache(self):
        locators: Dict[str, Dict] = {
            name: locator.to_dict() for name, locator in self.database.locators.items()
        }

        def load_image(path) -> Optional[str]:
            root = Path(self.database.path).parent
            try:
                with open(root / path, "rb") as infile:
                    return base64.b64encode(infile.read()).decode()
            except FileNotFoundError:
                self.logger.warning("Failed to read image: %s", path)
                return ""

        for locator in locators.values():
            if locator["type"] == "browser" and locator.get("screenshot"):
                locator["screenshot"] = load_image(locator["screenshot"])
            elif locator["type"] == "image":
                locator["value"] = load_image(locator["path"])

        self.cache = locators

    def get(self, name):
        if name not in self.database.locators:
            raise KeyError(f"No locator with name: {name}")

        return self.cache[name]

    def pop(self, name):
        if name not in self.database.locators:
            raise KeyError(f"No locator with name: {name}")

        del self.database.locators[name]
        return self.cache.pop(name)

    def load(self):
        self.database.load()
        self._create_cache()

    def save(self):
        # Filter out stored source images, if they exist
        root = Path(self.database.path).parent
        for locator in self.database.locators.values():
            if not isinstance(locator, ImageTemplate):
                continue
            if not getattr(locator, "source", None):
                continue
            try:
                Path(root / str(locator.source)).unlink()
                self.logger.info("Removed source image: %s", locator.source)
                locator.source = None
            except FileNotFoundError:
                pass

        self.database.save()
        self._create_cache()

    def list(self):
        """Convert current database to JSON."""
        if self.database.error:
            message = self.database.error[0] % self.database.error[1]
            raise DatabaseError(message)

        return self.cache

    def update(self, name, fields):
        """Update locator entries."""
        if not name.strip():
            raise DatabaseError("Empty/missing locator name")

        stored = self.database.locators.get(name, {})
        cached = self.cache.get(name, {})

        def required(src, dst=None):
            dst = dst or src
            path = getattr(stored, dst, None)
            if fields[src] != cached.get(src):
                fields[dst] = self._save_image(name, dst, fields[src], path)
            else:
                fields[dst] = path

        def optional(src, dst=None):
            dst = dst or src
            val = fields.get(src)
            path = getattr(stored, dst, None)
            if val:
                if val != cached.get(src):
                    fields[dst] = self._save_image(name, dst, val, path)
                else:
                    fields[dst] = path
            elif path:
                try:
                    Path(path).unlink()
                except FileNotFoundError:
                    pass

        if fields["type"] == "browser":
            optional("screenshot")
        elif fields["type"] == "image":
            required("value", "path")

        try:
            self.database.locators[name] = Locator.from_dict(fields)
        except ValueError as err:
            raise DatabaseError from err

        self.save()
        return self.list()

    def delete(self, name):
        """Delete locator entries."""
        if name not in self.database.locators:
            raise DatabaseError(f"Unknown locator: {name}")

        locator = self.database.locators.pop(name)

        def remove(key):
            path = getattr(locator, key, None)
            if not path:
                return
            try:
                Path(path).unlink()
            except FileNotFoundError:
                pass

        if isinstance(locator, BrowserDOM):
            remove("screenshot")
        elif isinstance(locator, ImageTemplate):
            remove("path")

        self.save()
        return self.list()

    def _save_image(self, name, field, content, path=None):
        """Save base64 image into a file."""
        root = Path(self.database.path).parent
        images = root / ".images"

        if path is not None and str(path).strip():
            path = root / path
        else:
            path = images / f"{sanitize_name(name)}-{field}.png"

        data = base64.b64decode(content)

        os.makedirs(images, exist_ok=True)
        with open(path, "wb") as outfile:
            outfile.write(data)

        return str(path.relative_to(root))



import base64
import logging
from dataclasses import dataclass
from io import BytesIO
from typing import List, Dict, Union, Tuple
from queue import Queue, Empty

import uiautomation as auto
import win32gui  # pylint: disable=import-error
import win32api  # pylint: disable=import-error
from PIL import ImageGrab
from RPA.core.geometry import Region
from RPA.core.windows import ElementInspector, RecordElement, WindowsElements
from pynput_robocorp.mouse import Listener

from inspector.windows.base import APP_WINDOW_TITLE


LOCATOR_VERSION = 1.0

PICK_TIMEOUT = 60

COLOR_RED = win32api.RGB(255, 0, 0)


WindowsLocatorProperties = Dict[str, str]
WindowsLocator = Dict[str, WindowsLocatorProperties]
MatchedWindowsLocator = Dict[str, Union[str, float]]
MatchedWindowsLocators = Dict[str, List[MatchedWindowsLocator]]
OpenWindow = Dict[str, str]
OpenWindows = List[OpenWindow]


@dataclass
class Rectangle(Region):
    @classmethod
    def from_element(cls, element):
        left = int(element.xcenter - (element.width / 2))
        right = int(element.xcenter + (element.width / 2))
        bottom = int(element.ycenter + (element.height / 2))
        top = int(element.ycenter - (element.height / 2))
        return cls(left, top, right, bottom)

    @classmethod
    def from_control(cls, control: auto.Control):
        left = int(control.BoundingRectangle.left)
        top = int(control.BoundingRectangle.top)
        right = int(control.BoundingRectangle.right)
        bottom = int(control.BoundingRectangle.bottom)
        return cls(left, top, right, bottom)

    def get_screenshot(self) -> str:
        image = ImageGrab.grab(self.as_tuple())
        jpeg_image_buffer = BytesIO()
        image.save(jpeg_image_buffer, format="JPEG")
        return base64.b64encode(jpeg_image_buffer.getvalue()).decode("utf-8")

    def get_borders(self, thickness: int) -> List[Tuple[int, int, int, int]]:
        """Get one pixel rectangles as tuples from rectangle coordinates"""
        rect = self.as_tuple()
        return [
            (rect[0] + i, rect[1] + i, rect[2] - i, rect[3] - i)
            for i in range(thickness)
        ]


class WindowMismatch(Exception):
    pass


class WindowsDriver:
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self._windows_elements = WindowsElements()
        self._active_window = ""
        self.queue = Queue()

    def _construct_locator(
        self,
        element: RecordElement,
    ) -> WindowsLocator:
        locator_path: str = element["locator"]
        control: auto.Control = element["control"]
        window_name: str = element["top"]
        if window_name != self._active_window:
            raise WindowMismatch("Non active window clicked")

        name = control.Name or "WindowsLocator"

        properties = {
            "window": self._active_window,
            "strategy": "WindowsLocator",
            "value": locator_path,
            "version": LOCATOR_VERSION,
        }

        if hasattr(control, "BoundingRectangle"):
            try:
                properties["screenshot"] = Rectangle.from_control(
                    control
                ).get_screenshot()
            except AttributeError as err:
                self.logger.error("Failed to get locator screenshot=%r", err)

        return {name: properties}

    def _on_click(
        self,
        x_coord: int,
        y_coord: int,
        element: auto.Control,  # pylint: disable=unused-argument
        pressed: bool,
    ) -> None:
        self.logger.debug("Element clicked (%s) at=%s,%s", pressed, x_coord, y_coord)

        locator = {}

        try:
            elements: List[RecordElement] = []
            ElementInspector.inspect_element(action="", recording=elements)
            locator = self._construct_locator(elements[0])
        except (ValueError, IndexError) as err:
            self.logger.error(err)
        except WindowMismatch as err:
            self.logger.info(err)
            return

        self.queue.put(locator)

    @classmethod
    def _find_all_top_level_window(cls) -> List[auto.Control]:
        return auto.GetRootControl().GetChildren()

    def _focus_active_window(self, active_window: str) -> None:
        self._active_window = active_window
        windows = self._find_all_top_level_window()
        for window in windows:
            if hasattr(window, "Name") and window.Name == self._active_window:
                if hasattr(window, "SwitchToThisWindow"):
                    window.SwitchToThisWindow()
                if hasattr(window, "SetFocus"):
                    window.SetFocus()
                break
        else:
            self.logger.error("Could not focus window=%s", active_window)

    def listen(self, active_window: str):
        self.logger.info("Start listening for mouse click in window=%s", active_window)
        self._focus_active_window(active_window)
        locator = {}
        listener = Listener(on_click=self._on_click)
        try:
            listener.start()
            locator = self.queue.get(block=True, timeout=PICK_TIMEOUT)
        except Empty:
            self.logger.info("Locator selection timeout")
        finally:
            listener.stop()
        return locator

    def validate(self, window: str, value: str) -> MatchedWindowsLocators:
        self.logger.info("Validate=%s, %s", window, value)
        locators: MatchedWindowsLocators = {"matches": []}
        self._focus_active_window(window)
        root = self._windows_elements.get_element(f"name:{window}")
        element = self._windows_elements.get_element(value, 8, root)
        if element and hasattr(element, "name") and hasattr(element, "locator"):
            locators["matches"].append(
                {
                    "window": window,
                    "name": element.name,
                    "value": element.locator,
                    "version": LOCATOR_VERSION,
                    "screenshot": Rectangle.from_element(element).get_screenshot(),
                }
            )
        return locators

    def list_windows(self) -> OpenWindows:
        self.logger.info("Get window info")
        windows = []
        try:
            windows = self._windows_elements.list_windows(icons=True)
        except TypeError as err:
            # If windows is closed while listing the windows library
            # will raise an exception.
            # The fix need to be done in the windows library side,
            # but will add this catch here until the fix is done
            self.logger.error("Failed to list windows: %s", str(err))
        returned_windows = []
        for window in windows:
            if window["title"] and window["title"] != APP_WINDOW_TITLE:
                win = {"title": window["title"]}
                if window["icon"]:
                    win["icon"] = window["icon"]
                returned_windows.append(win)
        return returned_windows

    def _draw_borders(self, rect: Rectangle, color: win32api.RGB, thickness: int = 5):
        try:
            full_screen_context = win32gui.GetDC(0)
            brush = win32gui.CreateSolidBrush(color)
            win32gui.SelectObject(full_screen_context, brush)
            frames = rect.get_borders(thickness)
            for frame in frames:
                win32gui.FrameRect(full_screen_context, frame, brush)
        except AttributeError as error:
            self.logger.error("Could not draw border=%s", error)

    def focus(self, window: str, locator: str) -> None:
        self.logger.info("Focus element=%s in window=%s", locator, window)
        self._focus_active_window(window)
        root = self._windows_elements.get_element(f"name:{window}")
        element = self._windows_elements.get_element(locator, 8, root)
        rect = Rectangle.from_element(element)
        self._draw_borders(rect, COLOR_RED)



import logging
import queue
import threading
from urllib.parse import quote

import requests
from requests.packages.urllib3.util.retry import Retry  # pylint: disable=import-error
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException
from inspector.metric import Metric


class TimeoutHTTPAdapter(HTTPAdapter):
    """Mountable timeout adapter for http(s) calls:
    https://github.com/psf/requests/issues/3070#issuecomment-205070203
    """

    DEFAULT_TIMEOUT = 5  # seconds

    def __init__(self, *args, **kwargs):
        if "timeout" in kwargs:
            self.timeout = kwargs.pop("timeout")
        else:
            self.timeout = self.DEFAULT_TIMEOUT
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):  # pylint: disable=arguments-differ
        if "timeout" not in kwargs:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


class Telemetry:  # pylint: disable=too-few-public-methods
    TIMEOUT = 1.5
    RETRIES = 2
    BASE_URL = "https://telemetry.robocorp.com/metric-v1/"

    def __init__(self, debug=False):
        self._logger = logging.getLogger(__name__)
        self._debug = debug
        self._thread = threading.Thread(target=self._run_worker)
        self._queue = queue.Queue()
        self._session = None

    def start_worker(self):
        self._thread.daemon = True
        self._thread.start()

    def send(self, metric: Metric):
        if not self._session:
            self._logger.debug("Telemetry not initialized")
            return

        if self._debug:
            self._logger.info("Suppressing telemetry: %s", metric)
            return

        self._queue.put(metric)

    def _run_worker(self):
        self._logger.debug("Starting telemetry worker")

        try:
            self._session = requests.Session()
            self._set_adapters()
            self._set_hooks()
        except RequestException as exc:
            self._logger.warning("Cannot initialize telemetry: %s", exc)
            self._session = None
            return

        while True:
            metric = self._queue.get()
            url = self.BASE_URL + quote(
                f"{metric.type}/"
                + f"{metric.timestamp}/"
                + f"{metric.instance_id}/"
                + f"{metric.key}/"
                + f"{metric.value}"
            )

            try:
                self._session.put(url)
                self._logger.info("Sent telemetry: %s", metric)
            except RequestException as error:
                self._logger.warning("Cannot send telemetry: %s", error)

    def _set_adapters(self):
        timeout = TimeoutHTTPAdapter(timeout=Telemetry.TIMEOUT)
        self._session.mount("https://", timeout)
        self._session.mount("http://", timeout)

        retry = HTTPAdapter(max_retries=self._retry_strategy())
        self._session.mount("https://", retry)
        self._session.mount("http://", retry)

    def _set_hooks(self):
        # pylint: disable=unused-argument
        def assert_status_hook(response, *args, **kwargs):
            response.raise_for_status()

        self._session.hooks["response"] = [assert_status_hook]

    def _retry_strategy(self) -> Retry:
        return Retry(
            total=self.RETRIES,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
            method_whitelist=["PUT"],
        )



import json
import subprocess
import sys
import time
from pathlib import Path

from inspector.windows.base import Bridge, Window, traceback
from inspector.windows.mixin import DatabaseMixin


def _base64_to_image(img):
    # pylint: disable=import-outside-toplevel
    from RPA.recognition import base64_to_image  # type: ignore

    return base64_to_image(img)


class ImageBridge(DatabaseMixin, Bridge):
    """Javascript API bridge for image template locators."""

    @traceback
    def pick(self, confidence=None):
        self.logger.info("Starting interactive picker")
        cmd = [
            sys.executable,
            str(Path(__file__).parent / ".." / "snipping_tool.py"),
        ]

        stdout = subprocess.check_output(cmd)
        if not stdout:
            raise RuntimeError("Picker closed abruptly")

        try:
            result = json.loads(stdout)
        except ValueError as err:
            raise RuntimeError(f"Malformed response from picker: {err}") from err

        if "error" in result:
            raise RuntimeError(result["error"])

        needle = _base64_to_image(result["image"])

        # TODO: No matches found if matching immediately, possibly
        # related to snipping tool closing.
        time.sleep(0.5)
        matches = self._find_matches(needle, confidence)

        return {"image": result["image"], "matches": [str(m) for m in matches]}

    @traceback
    def validate(
        self,
        value,
        confidence=None,
    ):
        needle = _base64_to_image(value)
        matches = self._find_matches(needle, confidence)

        # TODO: Replace with screenshots of matches?
        return {"matches": [str(m) for m in matches]}

    def _find_matches(self, needle, confidence=None):
        # pylint: disable=import-outside-toplevel
        import mss  # type: ignore
        from PIL import Image  # type: ignore
        from RPA.recognition import templates  # type: ignore
        from RPA.recognition.templates import (  # type: ignore
            ImageNotFoundError,
            DEFAULT_CONFIDENCE,
        )

        if confidence is None:
            confidence = DEFAULT_CONFIDENCE

        with mss.mss() as sct:
            monitor = sct.monitors[0]
            screenshot = sct.grab(monitor)
            haystack = Image.frombytes(
                "RGB", screenshot.size, screenshot.bgra, "raw", "BGRX"
            )

        try:
            self.logger.info("Finding template matches (confidence: %s)", confidence)
            matches = templates.find(haystack, needle, confidence=confidence)
        except ImageNotFoundError as err:
            self.logger.debug(str(err))
            matches = []

        self.logger.info(
            "Found %d match%s", len(matches), "" if len(matches) == 1 else "es"
        )

        return matches

    @traceback
    def minimize(self):
        self.logger.debug("Minimizing windows")
        for window in reversed(self.ctx.windows):
            if window.is_valid:
                window.minimize()

    @traceback
    def restore(self):
        self.logger.debug("Restoring windows")
        for window in self.ctx.windows:
            if window.is_valid:
                window.restore()


class ImageWindow(Window):
    BRIDGE = ImageBridge
    DEFAULTS = {
        "url": "image.html",
        "width": 560,
        "height": 720,
    }



import time
import uuid
from math import floor
from typing import Optional

from inspector import __version__


# TODO: Change metrics into dataclasses(?)
# pylint: disable=too-few-public-methods
class Metric:
    """Baseclass for a metric."""

    INSTANCE_ID: Optional[str] = None
    KEY: str = "inspector."

    def __init__(self, value: str):
        self.timestamp = str(int(floor(time.time() * 1000)))
        self.type = "test.inspector.dev"
        self.key = Metric.KEY + "base"
        self.value = value

    def __str__(self):
        return (
            f"{{id: {self.instance_id}, "
            + f"timestamp: {self.timestamp}, "
            + f"type: {self.type}, "
            + f"key: {self.key}, "
            + f"value: {self.value}}}"
        )

    @property
    def instance_id(self):
        if Metric.INSTANCE_ID is None:
            Metric.INSTANCE_ID = str(uuid.uuid4())
        return Metric.INSTANCE_ID


class MetricStart(Metric):
    """Class for a Start metric entry."""

    def __init__(self):
        super().__init__(value=__version__)
        self.key = Metric.KEY + "start"
